# -*- coding: utf-8 -*-

"""

tests.test_elements

Unit test the elements module (ie. Markdown source generation)


Copyright (C) 2024 Rainer Schwarzbach

This file is part of smdg.

smdg is free software: you can redistribute it and/or modify
it under the terms of the MIT License.

smdg is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
See the LICENSE file for more details.

"""


from unittest import TestCase

from smdg import elements
from smdg import html_generator as hg


EXPECTED_BQ_SIMPLE = "\n".join(("> line 1", "> line 2", "> line 3"))

# Examples from
# <https://daringfireball.net/projects/markdown/syntax#blockquote>

EXPECTED_RAW_HTML_EXAMPLE = """
This is a regular paragraph.

<table>
    <tr>
        <td>Foo</td>
    </tr>
</table>

This is another regular paragraph.
""".strip()

EXPECTED_SETEXT_HEADERS = """
This is an H1
=============

This is an H2
-------------
""".strip()

EXPECTED_ATX_HEADERS = """
# This is an H1

## This is an H2

###### This is an H6
""".strip()

EXPECTED_BQ_2_PARA = """
> This is a blockquote with two paragraphs. Lorem ipsum dolor sit amet,
> consectetuer adipiscing elit. Aliquam hendrerit mi posuere lectus.
> Vestibulum enim wisi, viverra nec, fringilla in, laoreet vitae, risus.
>
> Donec sit amet nisl. Aliquam semper ipsum sit amet velit. Suspendisse
> id sem consectetuer libero luctus adipiscing.
""".strip()

EXPECTED_BQ_NESTED = """
> This is the first level of quoting.
>
> > This is nested blockquote.
>
> Back to the first level.
""".strip()

EXPECTED_BQ_CONTAINER = """
> ## This is a header.
>
> 1.  This is the first list item.
> 2.  This is the second list item.
>
> Here's some example code:
>
>     return shell_exec("echo $input | $markdown_script");
""".strip()

EXPECTED_UNORDERED_LIST = """
*   Red
*   Green
*   Blue
""".strip()

EXPECTED_UNORDERED_LIST_HANGING_INDENTS = """
*   Lorem ipsum dolor sit amet, consectetuer adipiscing elit.
    Aliquam hendrerit mi posuere lectus. Vestibulum enim wisi,
    viverra nec, fringilla in, laoreet vitae, risus.
*   Donec sit amet nisl. Aliquam semper ipsum sit amet velit.
    Suspendisse id sem consectetuer libero luctus adipiscing.
""".strip()

EXPECTED_UNORDERED_LIST_OF_PARAGRAPHS = """
*   Bird

*   Magic
""".strip()

EXPECTED_UNORDERED_LIST_WITH_BLOCKQUOTE = """
*   A list item with a blockquote:

    > This is a blockquote
    > inside a list item.
""".strip()

EXPECTED_UNORDERED_LIST_WITH_CODEBLOCK = """
*   A list item with a code block:

        <code goes here>
""".strip()

EXPECTED_ORDERED_LIST = """
1.  Bird
2.  McHale
3.  Parish
""".strip()

EXPECTED_ORDERED_LIST_WITH_PARAGRAPHS = """
1.  This is a list item with two paragraphs. Lorem ipsum dolor
    sit amet, consectetuer adipiscing elit. Aliquam hendrerit
    mi posuere lectus.

    Vestibulum enim wisi, viverra nec, fringilla in, laoreet
    vitae, risus. Donec sit amet nisl. Aliquam semper ipsum
    sit amet velit.

2.  Suspendisse id sem consectetuer libero luctus adipiscing.
""".strip()

EXPECTED_CODEBLOCK = """    Here is an example of AppleScript:

        tell application "Foo"
            beep
        end tell
""".rstrip()

EXPECTED_CODEBLOCK_HTML = """    <div class="footer">
        &copy; 2004 Foo Corporation
    </div>
""".rstrip()

EXPECTED_HR_3_DASHES = "---"
EXPECTED_HR_7_ASTERISKS = "*******"

EXPECTED_LABEL_WITH_TITLE = '[foo]: http://example.com/ "Optional Title Here"'

EXPECTED_LABEL_WITH_LONGISH_URL_AND_TITLE = """
[id]: http://example.com/longish/path/to/resource/here
    "Optional Title Here"
""".strip()

EXPECTED_LABEL_WITH_PARENTHESIZED_TITLE = (
    """[foo]: http://example.com/ (Optional "Title" 'Here')"""
)
EXPECTED_LABEL_WITHOUT_TITLE = "[Daring Fireball]: http://daringfireball.net/"
EXPECTED_LABEL_REPR_WITH_TITLE = (
    "Label('foo', 'http://example.com/', title='Optional Title Here')"
)

EXPECTED_LINK_INLINE = "[example](http://example.com/)"
EXPECTED_LINK_REPR_INLINE = (
    "Link('example', ref=None, url='http://example.com/', title=None)"
)
EXPECTED_LINK_INLINE_WITH_TITLE = (
    """[example](http://example.com/ "Optional Title Here")"""
)
EXPECTED_LINK_REPR_INLINE_WITH_TITLE = (
    "Link('example', ref=None, url='http://example.com/',"
    " title='Optional Title Here')"
)
EXPECTED_LINK_REFERENCE = "[example][foo]"
EXPECTED_LINK_REPR_REFERENCE = (
    "Link('example', ref='foo', url=None, title=None)"
)

EXPECTED_IMAGE_INLINE = "![example](http://example.com/)"
EXPECTED_IMAGE_REPR_INLINE = (
    "Image('example', ref=None, url='http://example.com/', title=None)"
)
EXPECTED_IMAGE_INLINE_WITH_TITLE = (
    """![example](http://example.com/ "Optional Title Here")"""
)
EXPECTED_IMAGE_REPR_INLINE_WITH_TITLE = (
    "Image('example', ref=None, url='http://example.com/',"
    " title='Optional Title Here')"
)
EXPECTED_IMAGE_REFERENCE = "![example][foo]"
EXPECTED_IMAGE_REPR_REFERENCE = (
    "Image('example', ref='foo', url=None, title=None)"
)

EXPECTED_BOLD_TEXT = "**bold example**"
EXPECTED_ITALICS_INSIDE_BOLD_TEXT = "**bold _and italics_ example**"
EXPECTED_ITALICS = "_italics example_"
EXPECTED_BOLD_TEXT_INSIDE_ITALICS = "_italics **and bold** example_"

EXPECTED_CODESPAN_SIMPLE = "`<code_example>`"
EXPECTED_CODESPAN_WITH_BACKTICKS = "````md ```literal``` in code span````"
EXPECTED_CODESPAN_WITH_ADDITIONAL_SPACING = (
    "`` `nested code` with additional spacing ``"
)

EXPECTED_AUTOLINK_SIMPLE = "<email@example.com>"
EXPECTED_AUTOLINK_REPR = "AutoLink('email@example.com')"


class BaseElement(TestCase):
    """Test the RawHTMLBlock class"""

    maxDiff = None

    def test_str_simple(self) -> None:
        """test the string output"""
        b_elem = elements.BaseElement("This _is_ a base element.")
        self.assertEqual(
            str(b_elem),
            "This \\_is\\_ a base element.",
        )

    def test_repr_simple(self) -> None:
        """test the representation"""
        b_elem = elements.BaseElement("This _is_ a base element.")
        self.assertEqual(
            repr(b_elem),
            "BaseElement('This _is_ a base element.')",
        )


class InlineElement(TestCase):
    """Test the InlineElement class"""

    def test_init(self) -> None:
        """test the LF check"""
        self.assertRaisesRegex(
            ValueError,
            "^Inline elements must not contain line feeds$",
            elements.InlineElement,
            "more\nthan\na single line",
        )
        b_elem = elements.BaseElement("This _is_ a base element.")
        self.assertEqual(
            str(b_elem),
            "This \\_is\\_ a base element.",
        )


class RawHTMLBlock(TestCase):
    """Test the RawHTMLBlock class"""

    maxDiff = None

    def test_str_simple(self) -> None:
        """test the string representation"""
        paragraph_1 = elements.Paragraph("This is a regular paragraph.")
        raw_html_block = elements.RawHTMLBlock(
            hg.Table(hg.TableRow(hg.TableCell("Foo")))
        )
        paragraph_2 = elements.Paragraph("This is another regular paragraph.")
        self.assertEqual(
            elements.render(paragraph_1, raw_html_block, paragraph_2),
            EXPECTED_RAW_HTML_EXAMPLE,
        )


class Header(TestCase):
    """Test the Header class"""

    maxDiff = None

    def test_setext_style(self) -> None:
        """test the setext style string representation"""
        header_1 = elements.Header(1, "This is an H1", setext=True)
        header_2 = elements.Header(2, "This is an H2", setext=True)
        self.assertEqual(
            elements.render(header_1, header_2),
            EXPECTED_SETEXT_HEADERS,
        )

    def test_atx_style(self) -> None:
        """test the setext style string representation"""
        header_1 = elements.Header(1, "This is an H1")
        header_2 = elements.Header(2, "This is an H2")
        header_6 = elements.Header(6, "This is an H6")
        self.assertEqual(
            elements.render(header_1, header_2, header_6),
            EXPECTED_ATX_HEADERS,
        )

    def test_str_nested(self) -> None:
        """test the string representation"""
        bq_elem = elements.BlockQuote(
            "This is the first level of quoting.",
            elements.BlockQuote("This is nested blockquote."),
            "Back to the first level.",
        )
        self.assertEqual(str(bq_elem), EXPECTED_BQ_NESTED)


class BlockQuote(TestCase):
    """Test the BlockQuote class"""

    maxDiff = None

    def test_simple(self) -> None:
        """simple blockquote"""
        bq_elem = elements.BlockQuote("line 1\nline 2\nline 3")
        self.assertEqual(str(bq_elem), EXPECTED_BQ_SIMPLE)

    def test_2_paragraphs(self) -> None:
        """Two-paragraphs example from the spec"""
        para_1 = elements.Paragraph(
            "This is a blockquote with two paragraphs."
            " Lorem ipsum dolor sit amet,",
            "consectetuer adipiscing elit."
            " Aliquam hendrerit mi posuere lectus.",
            "Vestibulum enim wisi, viverra nec,"
            " fringilla in, laoreet vitae, risus.",
        )
        para_2 = elements.Paragraph(
            "Donec sit amet nisl. Aliquam semper ipsum sit amet velit."
            " Suspendisse",
            "id sem consectetuer libero luctus adipiscing.",
        )
        bq_2_para = elements.BlockQuote(para_1, para_2)
        self.assertEqual(str(bq_2_para), EXPECTED_BQ_2_PARA)

    def test_nested(self) -> None:
        """nested blockquotes"""
        bq_elem = elements.BlockQuote(
            "This is the first level of quoting.",
            elements.BlockQuote("This is nested blockquote."),
            "Back to the first level.",
        )
        self.assertEqual(str(bq_elem), EXPECTED_BQ_NESTED)

    def test_containers(self) -> None:
        """container lements"""
        bq_elem = elements.BlockQuote(
            elements.Header(2, "This is a header."),
            elements.OrderedList(
                "This is the first list item.",
                "This is the second list item.",
            ),
            elements.Paragraph("Here's some example code:"),
            elements.CodeBlock(
                'return shell_exec("echo $input | $markdown_script");'
            ),
        )
        self.assertEqual(str(bq_elem), EXPECTED_BQ_CONTAINER)


class CodeBlock(TestCase):
    """Test the CodeBlock class"""

    maxDiff = None

    def test_applescript(self) -> None:
        """code block: apple script, from the spec"""
        cb_elem = elements.CodeBlock(
            "Here is an example of AppleScript:\n"
            "\n"
            '    tell application "Foo"\n'
            "        beep\n"
            "    end tell"
        )
        self.assertEqual(str(cb_elem), EXPECTED_CODEBLOCK)

    def test_html(self) -> None:
        """code block: HTML, from the spec"""
        cb_elem = elements.CodeBlock(
            '<div class="footer">',
            "    &copy; 2004 Foo Corporation",
            "</div>",
        )
        self.assertEqual(str(cb_elem), EXPECTED_CODEBLOCK_HTML)


class HorizontalRule(TestCase):
    """Test the HorizontalRule class"""

    maxDiff = None

    def test_default(self) -> None:
        """default horizontal rule with 3 dashes"""
        hr_elem = elements.HorizontalRule()
        self.assertEqual(str(hr_elem), EXPECTED_HR_3_DASHES)

    def test_7_asterisks(self) -> None:
        """horizontal rule with 7 asterisks"""
        hr_elem = elements.HorizontalRule(7, "*")
        self.assertEqual(str(hr_elem), EXPECTED_HR_7_ASTERISKS)


class UnorderedList(TestCase):
    """Test the UnorderedList class"""

    def test_simple(self) -> None:
        """unordered list with simple elements"""
        ul_elem = elements.UnorderedList("Red", "Green", "Blue")
        self.assertEqual(str(ul_elem), EXPECTED_UNORDERED_LIST)

    def test_hanging_indents(self) -> None:
        """unordered list with simple elements and hanging indents"""
        ul_elem = elements.UnorderedList(
            "Lorem ipsum dolor sit amet, consectetuer adipiscing elit.\n"
            "Aliquam hendrerit mi posuere lectus. Vestibulum enim wisi,\n"
            "viverra nec, fringilla in, laoreet vitae, risus.",
            "Donec sit amet nisl. Aliquam semper ipsum sit amet velit.\n"
            "Suspendisse id sem consectetuer libero luctus adipiscing.",
        )
        self.assertEqual(str(ul_elem), EXPECTED_UNORDERED_LIST_HANGING_INDENTS)

    def test_with_blockquote(self) -> None:
        """unordered list with a blockquote"""
        ul_elem = elements.UnorderedList(
            elements.ListItem(
                "A list item with a blockquote:",
                elements.BlockQuote(
                    "This is a blockquote",
                    "inside a list item.",
                ),
            )
        )
        self.assertEqual(str(ul_elem), EXPECTED_UNORDERED_LIST_WITH_BLOCKQUOTE)

    def test_with_codeblock(self) -> None:
        """unordered list with a code block"""
        ul_elem = elements.UnorderedList(
            elements.ListItem(
                "A list item with a code block:",
                elements.CodeBlock("<code goes here>"),
            )
        )
        self.assertEqual(str(ul_elem), EXPECTED_UNORDERED_LIST_WITH_CODEBLOCK)

    def test_with_parapraphs(self) -> None:
        """unordered list with pragraphs"""
        ul_elem = elements.UnorderedList(
            elements.Paragraph("Bird"),
            elements.Paragraph("Magic"),
        )
        self.assertEqual(str(ul_elem), EXPECTED_UNORDERED_LIST_OF_PARAGRAPHS)


class OrderedList(TestCase):
    """Test the OrderedList class"""

    def test_simple(self) -> None:
        """unordered list with simple elements"""
        ol_elem = elements.OrderedList("Bird", "McHale", "Parish")
        self.assertEqual(str(ol_elem), EXPECTED_ORDERED_LIST)

    def test_with_paragraphs(self) -> None:
        """unordered list with simple elements"""
        ol_elem = elements.OrderedList(
            elements.ListItem(
                elements.Paragraph(
                    "This is a list item with two paragraphs."
                    " Lorem ipsum dolor",
                    "sit amet, consectetuer adipiscing elit."
                    " Aliquam hendrerit",
                    "mi posuere lectus.",
                ),
                elements.Paragraph(
                    "Vestibulum enim wisi, viverra nec, fringilla in, laoreet",
                    "vitae, risus. Donec sit amet nisl. Aliquam semper ipsum",
                    "sit amet velit.",
                ),
                number=1,
            ),
            elements.ListItem(
                elements.Paragraph(
                    "Suspendisse id sem consectetuer libero luctus adipiscing."
                ),
                number=2,
            ),
        )
        self.assertEqual(str(ol_elem), EXPECTED_ORDERED_LIST_WITH_PARAGRAPHS)


class Label(TestCase):
    """Test the Label class"""

    def test_with_title(self) -> None:
        """Test a label with title"""
        with self.subTest("default: title in parentheses"):
            label = elements.Label(
                "foo", "http://example.com/", title="Optional Title Here"
            )
            self.assertEqual(str(label), EXPECTED_LABEL_WITH_TITLE)
        #
        with self.subTest("fallback: title in quotes"):
            label = elements.Label(
                "foo",
                "http://example.com/",
                title="""Optional "Title" 'Here'""",
            )
            self.assertEqual(
                str(label), EXPECTED_LABEL_WITH_PARENTHESIZED_TITLE
            )
        #

    def test_with_lonmgish_url_and_title(self) -> None:
        """Test a label with title"""
        label = elements.Label(
            "id",
            "http://example.com/longish/path/to/resource/here",
            title="Optional Title Here",
        )
        self.assertEqual(str(label), EXPECTED_LABEL_WITH_LONGISH_URL_AND_TITLE)

    def test_without_title(self) -> None:
        """Test a label without title"""
        label = elements.Label("Daring Fireball", "http://daringfireball.net/")
        self.assertEqual(str(label), EXPECTED_LABEL_WITHOUT_TITLE)

    def test_repr_with_title(self) -> None:
        """Test a label representation with title"""
        label = elements.Label(
            "foo", "http://example.com/", title="Optional Title Here"
        )
        self.assertEqual(repr(label), EXPECTED_LABEL_REPR_WITH_TITLE)


class Link(TestCase):
    """Test the Link class"""

    def test_inline(self) -> None:
        """inline style"""
        link = elements.Link("example", url="http://example.com/")
        self.assertEqual(str(link), EXPECTED_LINK_INLINE)

    def test_repr_inline(self) -> None:
        """representation of an inline style link"""
        link = elements.Link("example", url="http://example.com/")
        self.assertEqual(repr(link), EXPECTED_LINK_REPR_INLINE)

    def test_inline_with_title(self) -> None:
        """inline style"""
        link = elements.Link(
            "example", url="http://example.com/", title="Optional Title Here"
        )
        self.assertEqual(str(link), EXPECTED_LINK_INLINE_WITH_TITLE)

    def test_repr_inline_with_title(self) -> None:
        """representation of an inline style link"""
        link = elements.Link(
            "example", url="http://example.com/", title="Optional Title Here"
        )
        self.assertEqual(repr(link), EXPECTED_LINK_REPR_INLINE_WITH_TITLE)

    def test_reference(self) -> None:
        """reference style"""
        link = elements.Link("example", ref="foo")
        self.assertEqual(str(link), EXPECTED_LINK_REFERENCE)

    def test_repr_reference(self) -> None:
        """representation of an reference style link"""
        link = elements.Link("example", ref="foo")
        self.assertEqual(repr(link), EXPECTED_LINK_REPR_REFERENCE)


class Image(TestCase):
    """Test the Image class"""

    def test_inline(self) -> None:
        """inline style"""
        image = elements.Image("example", url="http://example.com/")
        self.assertEqual(str(image), EXPECTED_IMAGE_INLINE)

    def test_repr_inline(self) -> None:
        """representation of an inline style image"""
        image = elements.Image("example", url="http://example.com/")
        self.assertEqual(repr(image), EXPECTED_IMAGE_REPR_INLINE)

    def test_inline_with_title(self) -> None:
        """inline style"""
        image = elements.Image(
            "example", url="http://example.com/", title="Optional Title Here"
        )
        self.assertEqual(str(image), EXPECTED_IMAGE_INLINE_WITH_TITLE)

    def test_repr_inline_with_title(self) -> None:
        """representation of an inline style image"""
        image = elements.Image(
            "example", url="http://example.com/", title="Optional Title Here"
        )
        self.assertEqual(repr(image), EXPECTED_IMAGE_REPR_INLINE_WITH_TITLE)

    def test_reference(self) -> None:
        """reference style"""
        image = elements.Image("example", ref="foo")
        self.assertEqual(str(image), EXPECTED_IMAGE_REFERENCE)

    def test_repr_reference(self) -> None:
        """representation of an reference style image"""
        image = elements.Image("example", ref="foo")
        self.assertEqual(repr(image), EXPECTED_IMAGE_REPR_REFERENCE)


class Emphasis(TestCase):
    """Test the BoldText and ItalicText classes"""

    def test_simple_bold(self) -> None:
        """simple bold text"""
        bold_elem = elements.BoldText("bold example")
        self.assertEqual(str(bold_elem), EXPECTED_BOLD_TEXT)

    def test_italics_inside_bold(self) -> None:
        """italics nested in bold text"""
        bold_elem = elements.BoldText(
            elements.CompoundInlineElement(
                "bold ", elements.ItalicText("and italics"), " example"
            )
        )
        self.assertEqual(str(bold_elem), EXPECTED_ITALICS_INSIDE_BOLD_TEXT)

    def test_nested_bold(self) -> None:
        """bold text nested in bold text: error"""
        self.assertRaisesRegex(
            ValueError,
            "^Bold text must not be nested",
            elements.BoldText,
            elements.CompoundInlineElement(
                "bold ", elements.BoldText("and bold"), " example"
            ),
        )

    def test_simple_italics(self) -> None:
        """simple italics text"""
        italics_elem = elements.ItalicText("italics example")
        self.assertEqual(str(italics_elem), EXPECTED_ITALICS)

    def test_bold_inside_italics(self) -> None:
        """bold text nested in italics"""
        italics_elem = elements.ItalicText(
            elements.CompoundInlineElement(
                "italics ", elements.BoldText("and bold"), " example"
            )
        )
        self.assertEqual(str(italics_elem), EXPECTED_BOLD_TEXT_INSIDE_ITALICS)

    def test_empty_italics(self) -> None:
        """Initialization with empty contents: error"""
        self.assertRaisesRegex(
            ValueError,
            "^Empty italic text is not supported",
            elements.ItalicText,
            "",
        )


class CodeSpan(TestCase):
    """Test the CodeSpan class"""

    def test_simple(self) -> None:
        """simple code span"""
        cs_elem = elements.CodeSpan("<code_example>")
        self.assertEqual(str(cs_elem), EXPECTED_CODESPAN_SIMPLE)

    def test_with_backticks(self) -> None:
        """code span with backticks"""
        cs_elem = elements.CodeSpan("md ```literal``` in code span")
        self.assertEqual(str(cs_elem), EXPECTED_CODESPAN_WITH_BACKTICKS)

    def test_with_additional_spacing(self) -> None:
        """code span with automatically inserted additional soacing"""
        cs_elem = elements.CodeSpan("`nested code` with additional spacing")
        self.assertEqual(
            str(cs_elem), EXPECTED_CODESPAN_WITH_ADDITIONAL_SPACING
        )

    def test_contents_required(self) -> None:
        """Empty code blocks are not allowed"""
        self.assertRaisesRegex(
            ValueError,
            "^Empty code blocks are not supported",
            elements.CodeSpan,
            "",
        )

    def test_no_lf_allowed(self) -> None:
        """No line feed allowed in code span"""
        self.assertRaisesRegex(
            ValueError,
            "^Inline elements must not contain line feeds",
            elements.CodeSpan,
            "one line \n another line",
        )


class AutoLink(TestCase):
    """Test the AutoLink class"""

    def test_simple(self) -> None:
        """simple AutoLink"""
        al_elem = elements.AutoLink("email@example.com")
        self.assertEqual(str(al_elem), EXPECTED_AUTOLINK_SIMPLE)

    def test_repr(self) -> None:
        """simple AutoLink representation"""
        al_elem = elements.AutoLink("email@example.com")
        self.assertEqual(repr(al_elem), EXPECTED_AUTOLINK_REPR)


# vim: fileencoding=utf-8 ts=4 sts=4 sw=4 autoindent expandtab syntax=python:



# -*- coding: utf-8 -*-

"""

smdg.html_generator

Minimalistic HTML generation helper module

THIS MODULE IMPLEMENTS ONLY A SMALL HTML ELEMENTS SUBSET
AND IS NOT INTENDED FOR GENERAL PURPOSES.


Copyright (C) 2024 Rainer Schwarzbach

This file is part of smdg.

smdg is free software: you can redistribute it and/or modify
it under the terms of the MIT License.

smdg is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
See the LICENSE file for more details.

"""


import html

from typing import List, Union


#
# Constants
#


BLANK = " "
EMPTY = ""
LF = "\n"
DASH = "-"
UNDERLINE = "_"
INDENT = "    "


__all__ = ["Element", "BlockElement"]

#
# Classes
#


class Element:
    """HTML base element:
    <span> as generic iniline element
    """

    name = "span"
    empty = False
    collapse_contents = True
    may_contain_block_elements = False

    def __init__(self, *contents: Union[str, "Element"], **attributes) -> None:
        """Initialize element and attributes"""
        self._attributes = attributes
        if not self.may_contain_block_elements:
            for item in contents:
                if isinstance(item, BlockElement):
                    raise ValueError(
                        f"{self.__class__.__name__} instances"
                        f" (here: {self.name})"
                        " must not contain block elements"
                    )
                #
            #
        #
        self._contents = contents

    def __str__(self) -> str:
        """String representation"""
        if self.empty:
            return f"<{self.name}{self.format_attributes()} />"
        #
        return (
            f"<{self.name}{self.format_attributes()}>"
            f"{self.format_contents()}</{self.name}>"
        )

    def format_attributes(self) -> str:
        """Format the attributes"""
        attributes_list: List[str] = [EMPTY]
        for key, value in self._attributes.items():
            attr_name = key.replace(UNDERLINE, DASH)
            if value is None:
                attributes_list.append(attr_name)
            else:
                attributes_list.append(
                    f'{attr_name}="{html.escape(str(value))}"'
                )
            #
        #
        return BLANK.join(attributes_list)

    def format_contents(self) -> str:
        """Format the contents"""
        output_list: List[str] = [str(item) for item in self._contents]
        if self.collapse_contents:
            return EMPTY.join(output_list)
        #
        output_block = LF.join(output_list)
        output_list = [EMPTY]
        for line in output_block.splitlines():
            output_list.append(f"{INDENT}{line}")
        #
        output_list.append(EMPTY)
        return LF.join(output_list)


class BlockElement(Element):
    """<div> as generic block element"""

    name = "div"
    collapse_contents = False
    may_contain_block_elements = True


class Header(BlockElement):
    """<h*>"""

    name = "h*"
    collapse_contents = True
    may_contain_block_elements = False

    def __init__(
        self, level: int, *contents: Union[str, "Element"], **attributes
    ) -> None:
        """Initialize element and attributes"""
        if not 0 < level < 7:
            raise ValueError("level must be between 1 and 6")
        #
        self.name = f"h{level}"
        super().__init__(*contents, **attributes)


class Paragraph(BlockElement):
    """<p>"""

    name = "p"
    collapse_contents = True


class Table(BlockElement):
    """<table>"""

    name = "table"
    collapse_contents = False


class TableRow(BlockElement):
    """<tr>"""

    name = "tr"
    collapse_contents = False


class TableCell(Element):
    """<td>"""

    name = "td"


# vim: fileencoding=utf-8 ts=4 sts=4 sw=4 autoindent expandtab syntax=python:



# -*- coding: utf-8 -*-

"""

tests.test_strings

Unit test the strings module


Copyright (C) 2024 Rainer Schwarzbach

This file is part of smdg.

smdg is free software: you can redistribute it and/or modify
it under the terms of the MIT License.

smdg is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
See the LICENSE file for more details.

"""


from unittest import TestCase

from smdg import strings


class Functions(TestCase):
    """Test the module functions"""

    def test_sanitize_specials(self) -> None:
        """sanitize() function with the SPECIALS regex"""
        for source, expected_result in (
            ("*", "\\*"),
            ("\\", "\\\\"),
            ("_", "\\_"),
            ("[", "\\["),
            ("]", "\\]"),
        ):
            with self.subTest(source=source, expected_result=expected_result):
                self.assertEqual(
                    strings.sanitize(source),
                    strings.SafeString(expected_result),
                )
            #
        #


# vim: fileencoding=utf-8 ts=4 sts=4 sw=4 autoindent expandtab syntax=python:



# -*- coding: utf-8 -*-

"""

smdg.strings

Source string handling


Copyright (C) 2024 Rainer Schwarzbach

This file is part of smdg.

smdg is free software: you can redistribute it and/or modify
it under the terms of the MIT License.

smdg is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
See the LICENSE file for more details.

"""


# import logging

import re

from typing import List, Sequence, Union


#
# Constants
#


BLANK = " "
EQUALS_OP = "="
EMPTY = ""
LF = "\n"
BACKSLASH = "\\"
BACKTICK = "`"
ASTERISK = "*"
POUND = "#"
UNDERLINE = "_"
SQUARE_BRACKETS = "[]"
PARENTHESES = "()"

SPECIALS = f"{ASTERISK}{UNDERLINE}{BACKSLASH}{BACKTICK}{SQUARE_BRACKETS}"
PRX_SPECIALS = re.compile(f"[{re.escape(SPECIALS)}]")
PRX_ASTERISKS_ONLY = re.compile(f"{re.escape(ASTERISK)}")
PRX_PARENTHESES_ONLY = re.compile(f"{re.escape(PARENTHESES)}")
PRX_UNDERLINES_ONLY = re.compile(f"{UNDERLINE}")
BACKSLASH_ESCAPED_REPLACEMENT = f"{3 * BACKSLASH}g<0>"


#
# Classes
#


class SafeString:
    """A string declared as safe"""

    def __init__(self, source: Union[str, "SafeString"]) -> None:
        """Store the source"""
        if isinstance(source, str):
            self._original = source
        else:
            self._original = str(source)
        #

    def __repr__(self) -> str:
        """String representation"""
        return f"{self.__class__.__name__}({repr(self._original)})"

    def __str__(self) -> str:
        """Source string"""
        return self._original

    def __hash__(self) -> int:
        """Hash over the string representation"""
        return hash(self._original)

    def __len__(self) -> int:
        """Length"""
        return len(self._original)

    def __contains__(self, item) -> bool:
        """'in' implementation"""
        if isinstance(item, str):
            maybe_contained = item
        else:
            maybe_contained = str(item)
        #
        return maybe_contained in self._original

    def __mul__(self, other) -> "SafeString":
        """Multiplied other times"""
        if isinstance(other, int):
            return SafeString(self._original * other)
        #
        raise ValueError("Can only multiply with integers")

    __imul__ = __mul__
    __rmul__ = __mul__

    def __eq__(self, other) -> bool:
        """Equality check"""
        return isinstance(other, SafeString) and str(self) == str(other)

    def __add__(self, other) -> "SafeString":
        """Add other"""
        if isinstance(other, str):
            return SafeString(f"{self._original}{other}")
        #
        return SafeString(f"{self._original}{str(other)}")

    __iadd__ = __add__
    __radd__ = __add__

    def copy(self) -> "SafeString":
        """String representation"""
        return SafeString(self._original)

    def rstrip(self) -> "SafeString":
        """strip trailing whitespace"""
        return SafeString(self._original.rstrip())

    def join(self, parts: Sequence["SafeString"]) -> "SafeString":
        """Return a safe joined string"""
        return self.__class__(str(self).join(str(item) for item in parts))

    def splitlines(self) -> List["SafeString"]:
        """Return a list of lines"""
        return [SafeString(line) for line in str(self).splitlines()]


#
# Helper Functions
#


def sanitize(
    source: str,
    pattern: re.Pattern[str] = PRX_SPECIALS,
    replacement: str = BACKSLASH_ESCAPED_REPLACEMENT,
) -> SafeString:
    """Return source with various characters defused"""
    return SafeString(pattern.sub(replacement, source))


def declare_as_safe(source: str) -> SafeString:
    """Return source directly declared as safe"""
    return SafeString(source)


# vim: fileencoding=utf-8 ts=4 sts=4 sw=4 autoindent expandtab syntax=python:



# -*- coding: utf-8 -*-

"""

smdg.elements

Markdown elements


Copyright (C) 2024 Rainer Schwarzbach

This file is part of smdg.

smdg is free software: you can redistribute it and/or modify
it under the terms of the MIT License.

smdg is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
See the LICENSE file for more details.

"""


# import logging

from typing import List, Optional, Set, Tuple, Union

# local modules

from smdg import html_generator
from smdg import strings


#
# Constants
#


ARGS_JOINER = ", "

# SafeStrinmg instances for single characters
SAFE_LF = strings.declare_as_safe(strings.LF)
SAFE_BLANK = strings.declare_as_safe(strings.BLANK)
SAFE_DOUBLE_QUOTE = strings.declare_as_safe('"')
SAFE_SINGLE_QUOTE = strings.declare_as_safe("'")
SAFE_EMPTY = strings.declare_as_safe(strings.EMPTY)
SAFE_EXCLAMATION_MARK = strings.declare_as_safe("!")
SAFE_BACKTICK = strings.declare_as_safe(strings.BACKTICK)
SAFE_UNDERLINE = strings.declare_as_safe(strings.UNDERLINE)
SAFE_ASTERISK = strings.declare_as_safe(strings.ASTERISK)
SAFE_POUND = strings.declare_as_safe(strings.POUND)
# Blockquote prefix
SAFE_BLOCKQUOTE_PREFIX = strings.declare_as_safe("> ")
# Left and Right Angle Brackets
SAFE_LAB = strings.declare_as_safe("<")
SAFE_RAB = strings.declare_as_safe(">")
# Left and Right Square Brackets
SAFE_LSB = strings.declare_as_safe("[")
SAFE_RSB = strings.declare_as_safe("]")
# Left and Right Parentheses
SAFE_LP = strings.declare_as_safe("(")
SAFE_RP = strings.declare_as_safe(")")
# Indent
INDENT = 4 * strings.BLANK
SAFE_INDENT = strings.declare_as_safe(INDENT)

__all__ = [
    "BaseElement",
    "InlineElement",
    "CompoundElement",
    "CompoundInlineElement",
    "BlockElement",
    "Paragraph",
    "RawHTMLBlock",
]


def render(
    *elements_sequence: Union[
        str, "InlineElement", "BlockElement", strings.SafeString
    ],
    indent_level: int = 0,
) -> str:
    """Render a sequence of elements"""
    indent = strings.EMPTY
    if indent_level > 0:
        indent = INDENT * indent_level
    #
    output_lines: List[str] = []
    previous_element: Optional[
        Union[str, InlineElement, BlockElement, strings.SafeString]
    ] = None
    for element in elements_sequence:
        # ensure block elements are divided by blank lines
        if (
            isinstance(element, BlockElement) and previous_element is not None
        ) or (
            isinstance(previous_element, BlockElement)
            and not isinstance(element, BlockElement)
        ):
            output_lines.append(strings.EMPTY)
        #
        #
        for line in element.splitlines():
            output_lines.append(f"{indent}{str(line)}")
        #
        previous_element = element
    #
    return strings.LF.join(output_lines)


class BaseElement(strings.SafeString):
    """MarkDown base class"""

    def __init__(self, source: Union[str, strings.SafeString]) -> None:
        """Store the defused internal source"""
        self.original_source_repr = repr(source)
        if isinstance(source, str):
            defused = strings.sanitize(source)
        else:
            defused = source
        #
        super().__init__(defused)

    def __repr__(self) -> str:
        """String representation"""
        return f"{self.__class__.__name__}({self.original_source_repr})"


class InlineElement(BaseElement):
    """Inline element, must not contain newlines"""

    def __init__(self, source: Union[str, strings.SafeString]) -> None:
        """Store the defused internal source"""
        super().__init__(source)
        if strings.LF in self:
            raise ValueError("Inline elements must not contain line feeds")
        #


class CompoundElement(BaseElement):
    """Compound MarkDown Element"""

    joiner: strings.SafeString = SAFE_EMPTY

    def __init__(
        self,
        *sources: Union[str, strings.SafeString],
    ) -> None:
        """Store the parts in a list"""
        self._sources: List[BaseElement] = []
        for item in sources:
            if isinstance(item, BaseElement):
                self._sources.append(item)
            else:
                self._sources.append(BaseElement(item))
            #
        #
        super().__init__(self.joiner.join(self._sources))
        self.original_source_repr = ARGS_JOINER.join(
            repr(single_source) for single_source in sources
        )

    def flattened(self):
        """return an iterator over flattened contents"""
        for item in self._sources:
            try:
                yield from item.flattened()
            except AttributeError:
                yield item
            #
        #


class CompoundInlineElement(CompoundElement, InlineElement):
    """Compound inline element"""

    def __init__(
        self,
        *sources: Union[str, strings.SafeString],
    ) -> None:
        """Ensure only inline elements are stored"""
        components: List[InlineElement] = []
        for item in sources:
            if isinstance(item, InlineElement):
                components.append(item)
            else:
                components.append(InlineElement(item))
            #
        #
        super().__init__(*components)
        self.original_source_repr = ARGS_JOINER.join(
            repr(single_source) for single_source in sources
        )


class BaseEmphasis(CompoundInlineElement):
    """Inline container element for emphasized text, base class"""

    description = "Emphasized"
    escape_pattern = strings.PRX_SPECIALS
    forbidden_contents: Set[strings.SafeString] = set()

    def __init__(
        self,
        source: Union[str, BaseElement],
        delimiter: strings.SafeString = SAFE_EMPTY,
    ) -> None:
        """Ensure the source does not contain forbidden contents"""
        if len(source) < 1:
            raise ValueError(
                f"Empty {self.description.lower()} text is not supported"
            )
        #
        if isinstance(source, str):
            checked_source: strings.SafeString = strings.sanitize(
                source, pattern=self.escape_pattern
            )
        else:
            if isinstance(source, CompoundElement):
                for item in source.flattened():
                    if item in self.forbidden_contents:
                        raise ValueError(
                            f"{self.description} text must not be nested"
                        )
                    #
                #
            #
            checked_source = source
        #
        super().__init__(delimiter, checked_source, delimiter)
        self.original_source_repr = (
            f"{source!r}{ARGS_JOINER}delimiter={delimiter!r}"
        )


class ItalicText(BaseEmphasis):
    """Italic text element"""

    description = "Italic"
    escape_pattern = strings.PRX_UNDERLINES_ONLY
    forbidden_contents: Set[strings.SafeString] = {SAFE_UNDERLINE}

    def __init__(self, source: Union[str, BaseElement]) -> None:
        """Make sure the source contains neither unescaped underlines
        nor any newlines, and prohibit nested italic text
        """
        super().__init__(source, delimiter=SAFE_UNDERLINE)
        self.original_source_repr = repr(source)


class BoldText(BaseEmphasis):
    """Italic text element"""

    description = "Bold"
    escape_pattern = strings.PRX_ASTERISKS_ONLY
    forbidden_contents: Set[strings.SafeString] = {
        SAFE_ASTERISK,
        2 * SAFE_ASTERISK,
    }

    def __init__(self, source: Union[str, BaseElement]) -> None:
        """Make sure the source contains neither unescaped asterisks
        nor any newlines, and prohibit nested bold text
        """
        super().__init__(source, delimiter=2 * SAFE_ASTERISK)
        self.original_source_repr = repr(source)


class BlockElement(CompoundElement):
    """Block element"""

    joiner: strings.SafeString = SAFE_LF
    prefix: Optional[strings.SafeString] = None
    indent_level = 0

    def __init__(self, *sources: Union[str, strings.SafeString]) -> None:
        """Store the lines in a CompoundElement,
        sanitize strings if necessary
        """
        source_elements: List[strings.SafeString] = []
        for element in sources:
            if isinstance(element, strings.SafeString):
                source_elements.append(element)
            else:
                source_elements.append(strings.sanitize(element))
            #
        #
        for element in source_elements:
            # Forbid Raw HTML elements inside other Elements
            if isinstance(element, RawHTMLBlock):
                raise ValueError(
                    "RawHTMLBlock elements must not be contained"
                    " in other elements"
                )
            #
        #
        self._source_lines: List[InlineElement] = []
        for line in render(*source_elements).splitlines():
            self._source_lines.append(
                self._prefixed_line(strings.declare_as_safe(line))
            )
        #
        super().__init__(*self._source_lines)
        self.original_source_repr = ARGS_JOINER.join(
            repr(single_source) for single_source in sources
        )

    def _prefixed_line(
        self, line: Union[str, strings.SafeString]
    ) -> InlineElement:
        """Prefix the line if a prefix has been set"""
        if self.prefix is None:
            return InlineElement(line)
        #
        if len(line) < 1:
            return InlineElement(self.prefix.rstrip())
        #
        return CompoundInlineElement(self.prefix, line)


class RawHTMLBlock(BlockElement):
    """Raw HTML block"""

    def __init__(
        self, html_block_element: html_generator.BlockElement
    ) -> None:
        """Initialize with the HTML block 'as is'"""
        super().__init__(strings.declare_as_safe(str(html_block_element)))
        self.original_source_repr = repr(html_block_element)


class Paragraph(BlockElement):
    """Paragraph"""


class Header(BlockElement):
    """Header"""

    setext_underline: Tuple[str, str, str] = ("", "=", "-")

    def __init__(
        self,
        level: int,
        source: Union[str, strings.SafeString],
        setext: bool = False,
    ) -> None:
        """Initialize element and attributes"""
        if not 0 < level < 7:
            raise ValueError("level must be between 1 and 6")
        #
        if isinstance(source, str):
            source = strings.sanitize(source)
        #
        if strings.LF in str(source):
            raise ValueError("Headers must not contain line feeds")
        #
        contents: List[strings.SafeString] = []
        if setext and level < 3:
            contents.append(source)
            contents.append(
                strings.declare_as_safe(
                    self.setext_underline[level] * len(source)
                )
            )
        else:
            contents.append(
                CompoundInlineElement(SAFE_POUND * level, SAFE_BLANK, source)
            )
        #
        super().__init__(*contents)
        self.original_source_repr = (
            f"{level}{ARGS_JOINER}{source!r}" f"{ARGS_JOINER}setext={setext!r}"
        )


class HorizontalRule(BlockElement):
    """Horizontal rule"""

    allowed_characters = "*-_"

    def __init__(
        self,
        number: int = 3,
        character: str = "-",
    ) -> None:
        """Initialize element and attributes"""
        if number < 3:
            raise ValueError("number of characters must be >= 3")
        #
        if len(character) != 1:
            raise ValueError("exactly one character is required")
        #
        if character not in self.allowed_characters:
            raise ValueError(
                f"character must be one of {self.allowed_characters.split()}"
            )
        #
        super().__init__(strings.declare_as_safe(number * character))
        self.original_source_repr = (
            f"number={number}{ARGS_JOINER}character={character!r}"
        )


class BlockQuote(BlockElement):
    """BlockQuote"""

    prefix = SAFE_BLOCKQUOTE_PREFIX


class ListItem(CompoundElement):
    """Base class for list elements"""

    joiner: strings.SafeString = SAFE_LF

    def __init__(
        self,
        *sources: Union[str, strings.SafeString],
        number: Optional[int] = None,
    ) -> None:
        """Store the lines of the indented list element"""
        self.block_display = False
        self.source_elements: List[strings.SafeString] = []
        for element in sources:
            if isinstance(element, BlockElement):
                self.block_display = True
            #
            if isinstance(element, strings.SafeString):
                self.source_elements.append(element)
            else:
                self.source_elements.append(strings.sanitize(element))
            #
        #
        if number is None:
            # unordered list
            prefix = "*"
        else:
            prefix = f"{number}."
        #
        additional_indent = len(prefix) // 4
        indent_level = 1 + additional_indent
        indent_chars = 4 * indent_level
        list_item_prefix = f"{prefix:<{indent_chars}}"
        blank_prefix = strings.BLANK * indent_chars
        # render contents, indented
        first_line = True
        item_lines: List[strings.SafeString] = []
        for line in render(*self.source_elements).splitlines():
            if first_line:
                output_prefix = list_item_prefix
                first_line = False
            else:
                output_prefix = blank_prefix
            #
            current_line = line
            if line or output_prefix.strip():
                current_line = f"{output_prefix}{line}"
            #
            item_lines.append(strings.declare_as_safe(current_line))
        #
        super().__init__(*item_lines)
        self.original_source_repr = ARGS_JOINER.join(
            [repr(single_source) for single_source in sources]
            + [f"number={number!r}"]
        )


class BaseList(BlockElement):
    """Base class for list elements"""

    start_number: Optional[int] = None
    increment: int = 1

    def __init__(self, *sources: Union[str, strings.SafeString]) -> None:
        """Store the list items ..."""
        items: List[ListItem] = []
        self.block_items = False
        current_number: Optional[int] = self.start_number
        for element in sources:
            if isinstance(element, ListItem):
                current_item: ListItem = element
            else:
                current_item = ListItem(element, number=current_number)
            #
            if current_item.block_display:
                self.block_items = True
            #
            items.append(current_item)
            if isinstance(current_number, int):
                current_number += self.increment
            #
        #
        separator = strings.LF
        if self.block_items:
            separator = 2 * strings.LF
        #
        output_parts: List[str] = []
        first_item = True
        for single_item in items:
            if first_item:
                output_parts.append(str(single_item))
                first_item = False
            else:
                output_parts.append(f"{separator}{str(single_item)}")
            #
        #
        # logging.warning("Output parts: %r", output_parts)
        super().__init__(
            strings.declare_as_safe(strings.EMPTY.join(output_parts))
        )
        self.original_source_repr = ARGS_JOINER.join(
            repr(single_source) for single_source in sources
        )


class UnorderedList(BaseList):
    """Unordered list"""


class OrderedList(BaseList):
    """Ordered list"""

    def __init__(
        self, *sources: Union[str, strings.SafeString], start_number: int = 1
    ) -> None:
        """Store the list items ..."""
        if start_number < 1:
            raise ValueError(f"Invalid start number: {start_number}")
        #
        self.start_number = start_number
        super().__init__(*sources)
        self.original_source_repr = ARGS_JOINER.join(
            [repr(single_source) for single_source in sources]
            + [f"start_number={start_number!r}"]
        )


class CodeBlock(BlockElement):
    """Indented code block"""

    prefix = SAFE_INDENT

    def __init__(self, *sources: Union[str, strings.SafeString]) -> None:
        """Store the lines as-is"""
        source_elements: List[strings.SafeString] = []
        for element in sources:
            if isinstance(element, strings.SafeString):
                source_elements.append(element)
            else:
                source_elements.append(strings.declare_as_safe(element))
            #
        #
        super().__init__(*source_elements)
        self.original_source_repr = ARGS_JOINER.join(
            repr(single_source) for single_source in sources
        )


class Label(BlockElement):
    """Label for a reference-style link or image"""

    break_threshold = 65

    def __init__(
        self, label_id: str, url: str, title: Optional[str] = None
    ) -> None:
        """Initialize"""
        core_label = CompoundInlineElement(
            SAFE_LSB,
            strings.sanitize(label_id),
            SAFE_RSB,
            strings.declare_as_safe(": "),
            strings.declare_as_safe(url),
        )
        lines: List[CompoundInlineElement] = []
        if title is None:
            lines.append(core_label)
        else:
            add_line_break = (
                len(title) + len(label_id) + len(url) >= self.break_threshold
            )
            components: List[strings.SafeString] = []
            if add_line_break:
                components.append(SAFE_INDENT)
            else:
                components.append(SAFE_BLANK)
            #
            if '"' not in title:
                safe_title_start = safe_title_end = SAFE_DOUBLE_QUOTE
            elif "'" not in title:
                safe_title_start = safe_title_end = SAFE_SINGLE_QUOTE
            else:
                safe_title_start = SAFE_LP
                safe_title_end = SAFE_RP
            #
            components.extend(
                (
                    safe_title_start,
                    strings.sanitize(title),
                    safe_title_end,
                )
            )
            title_part = CompoundInlineElement(*components)
            if add_line_break:
                lines.extend((core_label, title_part))
            else:
                lines.append(CompoundInlineElement(core_label, title_part))
            #
        #
        # TODO: sanitize label_id and URL
        super().__init__(*lines)
        self.original_source_repr = (
            f"{label_id!r}{ARGS_JOINER}{url!r}{ARGS_JOINER}title={title!r}"
        )


class BaseLink(CompoundInlineElement):
    """Link and image base class"""

    prefix: Optional[strings.SafeString] = None

    def __init__(
        self,
        text: str,
        ref: Optional[str] = None,
        url: Optional[str] = None,
        title: Optional[str] = None,
    ) -> None:
        """Initialize"""
        if (url is None and ref is None) or (
            url is not None and ref is not None
        ):
            raise ValueError("Specify either url or ref")
        #
        components: List[strings.SafeString] = []
        if isinstance(self.prefix, strings.SafeString):
            components.append(self.prefix)
        #
        components.extend((SAFE_LSB, strings.sanitize(text), SAFE_RSB))
        if isinstance(ref, str):
            components.extend(
                (
                    SAFE_LSB,
                    strings.sanitize(ref),
                    SAFE_RSB,
                )
            )
        elif isinstance(url, str):
            # TODO: sanitize URL
            parenthesized_content: List[strings.SafeString] = [
                strings.declare_as_safe(url)
            ]
            if isinstance(title, str):
                if '"' in title:
                    raise ValueError("Title must not contain double quotes")
                #
                parenthesized_content.extend(
                    (
                        SAFE_BLANK,
                        SAFE_DOUBLE_QUOTE,
                        strings.sanitize(title),
                        SAFE_DOUBLE_QUOTE,
                    )
                )
            #
            components.extend(
                (
                    SAFE_LP,
                    *parenthesized_content,
                    SAFE_RP,
                )
            )
        #
        super().__init__(*components)
        self.original_source_repr = (
            f"{text!r}{ARGS_JOINER}ref={ref!r}{ARGS_JOINER}url={url!r}"
            f"{ARGS_JOINER}title={title!r}"
        )


class Link(BaseLink):
    """Hyperlink"""


class Image(BaseLink):
    """Image"""

    prefix: strings.SafeString = SAFE_EXCLAMATION_MARK


class CodeSpan(CompoundInlineElement):
    """Backticks quoted inline literal"""

    def __init__(self, source: Union[str, strings.SafeString]) -> None:
        """Declare the internal source as safe and surround with backticks"""
        if len(source) < 1:
            raise ValueError("Empty code blocks are not supported")
        #
        quote = SAFE_BACKTICK
        if isinstance(source, strings.SafeString):
            safe_source = source
        else:
            safe_source = strings.declare_as_safe(source)
        #
        # Increase number of quotes if required
        while quote in safe_source:
            quote += SAFE_BACKTICK
        #
        # Surround enclosed text with blanks if it starts or ends
        # with a backtick
        raw = str(safe_source)
        if any(raw[index] == strings.BACKTICK for index in (0, -1)):
            safe_source = SAFE_BLANK + safe_source + SAFE_BLANK
        #
        super().__init__(quote, safe_source, quote)
        self.original_source_repr = repr(source)


class AutoLink(CompoundInlineElement):
    """Backticks quoted inline literal"""

    def __init__(self, url: Union[str, strings.SafeString]) -> None:
        """Declare the internal source as safe and surround with backticks"""
        if len(url) < 1:
            raise ValueError("Empty urls are not supported")
        #
        # TODO: sanitize URL
        if isinstance(url, strings.SafeString):
            safe_url = url
        else:
            safe_url = strings.declare_as_safe(url)
        #
        super().__init__(SAFE_LAB, safe_url, SAFE_RAB)
        self.original_source_repr = repr(url)


# vim: fileencoding=utf-8 ts=4 sts=4 sw=4 autoindent expandtab syntax=python:
